from komora_syncer.models.komora.KomoraApi import KomoraApi
from komora_syncer.models.netbox.NbSite import NbSite
from komora_syncer.models.netbox.NbRegion import NbRegion
from komora_syncer.models.netbox.NbTenant import NbTenant
from komora_syncer.models.netbox.NbDevice import NbDevice

from komora_syncer.connections.NetboxConnection import NetboxConnection

import logging
logger = logging.getLogger(__name__)


class Synchronizer:
    def __init__(self):
        self.komora_api = KomoraApi()

        # Lists of komora sites
        self.sites = []

        # All location data from komora - RUIAN
        self.regions_all = []
        self.districs_all = []
        self.municipalities_all = []

        # Locations assigned to Sites in komora (will be created in Netbox)
        self.regions_assigned = []
        self.districts_assigned = []
        self.municipalities_assigned = []

    # pull all Regions and sites from komora
    def __prepare_all_regions(self):
        if not self.regions_all:
            self.regions_all = self.komora_api.get_regions()
        if not self.municipalities_all:
            self.municipalities_all = self.komora_api.get_municipalities()
        if not self.districs_all:
            self.districs_all = self.komora_api.get_districts()
        if not self.sites:
            self.sites = self.komora_api.get_sites()

    # Fill the assigned locations only
    def __prepare_assigned_regions(self):
        if not self.regions_all or not self.municipalities_all or not self.districs_all or not self.sites:
            self.__prepare_all_regions()

        # municipalities sets on sites - site.address exists
        municipalities_on_sites = [[site.address.municipalityName,
                                    site.address.districtName] for site in self.sites if site.address]
        self.municipalities_assigned = [muni for muni in self.municipalities_all if [
            muni.name, muni.districtName] in municipalities_on_sites]

        # get districts from step aboce
        districts_on_assigned_municipalities = set(
            municipality.districtName for municipality in self.municipalities_assigned)
        self.districts_assigned = [
            dis for dis in self.districs_all if dis.name in districts_on_assigned_municipalities]

        # get regions from step above
        regions_on_assigned_districts = set(
            district.regionName for district in self.districts_assigned)
        self.regions_assigned = [
            reg for reg in self.regions_all if reg.name in regions_on_assigned_districts]

        # get sites where address from ruian does not exists
        #   but region is assigned - foreign regions for example
        sites_without_address_with_region = [[site.regionId, site.regionName] for site in self.sites if not site.address and site.regionId]
        # get regions from sites without address and append them
        self.regions_assigned.extend([reg for reg in self.regions_all if [reg.id, reg.name] in sites_without_address_with_region])

    def sync_regions(self):
        if not self.regions_assigned or not self.municipalities_assigned or not self.districts_assigned or not self.sites:
            self.__prepare_assigned_regions()

        logger.info("Synchronizing regions")
        # sync regions
        for reg in self.regions_assigned:
            nb_region = NbRegion(reg, 'region')
            nb_region.synchronize()

        # sync Districts
        for dis in self.districts_assigned:
            nb_region = NbRegion(dis, 'district')
            nb_region.synchronize()

        # sync Municipatilies
        for muni in self.municipalities_assigned:
            nb_region = NbRegion(muni, 'municipality')
            nb_region.synchronize()

    def sync_sites(self):
        try:
            if not self.sites:
                self.sites = self.komora_api.get_sites()
        except Exception as e:
            raise e

        logger.info("Synchronizing sites")
        for site in self.sites:
            nb_site = NbSite(site)
            nb_site.synchronize()

    def sync_organizations(self):
        organizations = self.komora_api.get_organizations()

        logger.info("Synchronizing organizations")
        for organization in organizations:
            nb_tenant = NbTenant(organization)
            nb_tenant.synchronize()

    def sync_devices(self):
        try:
            logger.info("Posting devices")
            post_devices = NbDevice.get_nb_devices_data()
            self.komora_api.post_devices(post_devices)
        except Exception as e:
            raise e

        try:
            devices = self.komora_api.get_devices()
        except Exception as e:
            raise e

        logger.info("Synchronizing devices")
        for dev in devices:
            nb_dev = NbDevice(dev)
            nb_dev.synchronize()

        # TODO: Make method
        # Get all komora_id from devices at Komora system
        komora_ids = [dev.id for dev in devices]

        pynetbox = NetboxConnection()
        pynetbox.open()

        # filter devices, which are not stored in KOMORA
        devices_not_in_komora = pynetbox.connection.dcim.devices.filter(
            cf_komora_id__n=komora_ids)

        # Unset komora link to those devices
        devs_to_update = []
        for dev in devices_not_in_komora:
            dev.custom_fields["komora_id"] = None
            dev.custom_fields["komora_url"] = None
            devs_to_update.append(dev)

        # Update netbox DB
        try:
            if pynetbox.connection.dcim.devices.update(devs_to_update):
                logger.info(f"Devices links to Komora successfully unset")
            else:
                logger.info(f"No device found for unsetting links to Komora")
        except Exception as e:
            raise e



class Device():
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.description = kwargs.get('description')
        self.primaryIP4Id = kwargs.get('primaryIP4Id')
        self.primaryIP4Address = kwargs.get('primaryIP4Address')
        self.serialNumber = kwargs.get('serialNumber')
        self.locationId = kwargs.get('locationId')
        self.locationName = kwargs.get('locationName')
        self.siteId = kwargs.get('siteId')
        self.siteName = kwargs.get('siteName')
        self.organizationId = kwargs.get('organizationId')
        self.organizationName = kwargs.get('organizationName')
        self.rackId = kwargs.get('rackId')
        self.rackFace = kwargs.get('rackFace')
        self.validFrom = kwargs.get('validFrom')
        self.validTo = kwargs.get('validTo')
        self.isActive = kwargs.get('isActive')
        self.id = kwargs.get('id')

        self.netBoxId = kwargs.get('netBoxId')

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name



#from pynetbox import RequestError
from komora_syncer.models.netbox.NetboxBase import NetboxBase
from slugify import slugify
from komora_syncer.helpers.utils import build_tenant_name

import traceback
import logging
logger = logging.getLogger(__name__)


class NbTenant(NetboxBase):
    def __init__(self, komora_obj=None):
        NetboxBase.__init__(self)

        self.name = build_tenant_name(komora_obj.clientId, komora_obj.name)
        self.slug = slugify(self.name)
        self.description = komora_obj.vipNote

        # Komora returns description as space
        if self.description.isspace():
            self.description = ""
        self.custom_fields = {
            "client_id": komora_obj.clientId,
            "komora_id": komora_obj.id,
            "client_name": komora_obj.name
        }

        # TODO: Remove
        if komora_obj.isCustomer == True:
            self.description += ";isCustomer;"

        self.description = self.description.strip()
        self.api_object = None

    def find_or_create(self):
        self.find()
        if not self.api_object():
            self.create()

        return self.api_object

    def find(self):
        # 1. lookup by KOMORA_ID
        # 2. Lookup by name, if komora id is not preseted
        # - log a problem, when the name exists, but komora_id was not found

        komora_id = self.custom_fields.get("komora_id", None)

        if komora_id:
            try:
                netbox_tenant = self.netbox.connection.tenancy.tenants.get(
                    cf_komora_id=komora_id)
                if netbox_tenant:
                    self.api_object = netbox_tenant
            # TODO:
            #except RequestError as e:
            #    logger.error(f"Unable to complete request - {e}")
            #    logger.debug(e)
            #    raise(e)
            except Exception as e:
                logger.exception(f"Unable to find tenant by komora_id: {komora_id}")
                #raise(e)
                netbox_tenant = None

        if not netbox_tenant:
            try:
                netbox_tenant = self.netbox.connection.tenancy.tenants.get(
                    name__ie=self.name)

                if netbox_tenant:
                    logger.warning(
                        f"komora_id: {str(komora_id)} was not found, but Tenant {self.name} already exists")
                    self.api_object = netbox_tenant
            except Exception as e:
                logger.exception(f"Unable to find tenant by name: {self.name}")
                raise(e)

        return self.api_object

    def create(self):
        try:
            params = self.get_params()
            netbox_tenant = self.netbox.connection.tenancy.tenants.create(
                params)

            logger.info("Region: %s created sucessfully", self.name)
            self.api_object = netbox_tenant
        except Exception as e:
            # TODO: Univerzita Tomáše Bati ve Zlíně, Fakulta aplikované informatiky - exists multiple times
            logger.error(f"Unable to create netbox tenant: {self.name}\n{e}")
            logger.debug(f"{traceback.format_exc()}")


        return self.api_object

    def update(self, nb_tenant):
        try:
            if nb_tenant.update(self.get_params()):
                self.api_object = nb_tenant
                logger.info(f"Tenant: {self.name} updated successfuly")
        except Exception as e:
            logger.error(e)
            raise e

    def synchronize(self):
        tenant = self.find()

        if tenant:
            self.update(tenant)
        else:
            self.create()

    def get_params(self):
        params = {'name': self.name,
                  'slug': self.slug,
                  'description': self.description,
                  'custom_fields': self.custom_fields}

        return params



class Location():
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.description = kwargs.get('description')
        self.roomPin = kwargs.get('roomPin')
        self.parentId = kwargs.get('parentId')
        self.id = kwargs.get('id')

        if kwargs.get('locations'):
            self.locations = []
            for location in kwargs.get('locations'):
                self.locations.append(Location(**location))
        else:
            self.locations = kwargs.get('locations')



class Municipality():
    def __init__(self, **kwargs):
        self.name = kwargs.get('name')
        self.districtId = kwargs.get('districtId')
        self.districtName = kwargs.get('districtName')
        self.regionId = kwargs.get('regionId')
        self.regionName = kwargs.get('regionName')
        self.isActive = kwargs.get('isActive')
        self.id = kwargs.get('id')



class Organization():
    def __init__(self, **kwargs):
        self.id = kwargs.get('id')
        self.abraCustomerId = kwargs.get('abraCustomerId')
        self.addres = kwargs.get('address')
        self.hasActiveContract = kwargs.get('hasActiveContract')
        self.icoNumber = kwargs.get('icoNumber')
        self.isActual = kwargs.get('isActual')
        self.isCustomer = kwargs.get('isCustomer')
        self.name = kwargs.get('name')
        self.abraCode = kwargs.get('abraCode')
        self.xUcastnikId = kwargs.get('xUcastnikId')
        self.clientId = kwargs.get('clientId')
        self.updatedAt = kwargs.get('updatedAt')
        self.updatedBy = kwargs.get('updatedBy')
        self.isMember = kwargs.get('isMember')
        self.isCanceledMemberPart = kwargs.get('isCanceledMemberPart')
        self.masterOrganizationId = kwargs.get('masterOrganizationId')
        self.masterOrganizationAbraId = kwargs.get('masterOrganizationAbraId')
        self.masterOrganizationName = kwargs.get('masterOrganizationName')
        self.entityId = kwargs.get('entityId')
        self.isVip = kwargs.get('isVip')
        self.vipNote = kwargs.get('vipNote')
        self.organizationTypeId = kwargs.get('organizationTypeId')
        self.organizationTypeName = kwargs.get('organizationTypeName')



from komora_syncer.models.komora.location import Location
from komora_syncer.models.komora.address import Address
from komora_syncer.models.komora.site_contact import SiteContact


class Site():
    def __init__(self, **kwargs):
        self.name = (kwargs.get('name') or "").strip()
        self.fullName = (kwargs.get('fullName') or "").strip()
        self.description = (kwargs.get('description') or "").strip()
        self.facility = kwargs.get('facility')
        self.latitude = kwargs.get('latitude')
        self.longitude = kwargs.get('longitude')
        self.code = kwargs.get('code')
        self.typeId = kwargs.get('typeId')
        self.typeName = kwargs.get('typeName')
        self.isActive = kwargs.get('isActive')
        self.id = kwargs.get('id')

        self.organizationId = kwargs.get('organizationId')
        self.organizationName = kwargs.get('organizationName')

        if kwargs.get('address', None):
            self.address = Address(**kwargs.get('address'))
        else:
            self.address = kwargs.get('address')

        if kwargs.get('locations', None):
            self.locations = []
            for location in kwargs.get('locations'):
                self.locations.append(Location(**location))
        else:
            self.locations = kwargs.get('locations')

        if kwargs.get('contacts', None):
            self.contacts = []
            for contact in kwargs.get('contacts'):
                self.contacts.append(SiteContact(**contact))
        else:
            self.contacts = kwargs.get('contacts')

        self.flatten_locations = flatten_locations(
            self.locations) if self.locations else []

        self.regionId = kwargs.get('regionId')
        self.regionName = kwargs.get('regionName')

def flatten_locations(nested_locations):
    result = []

    def flat(nested_locations):
        for location in nested_locations:
            result.append(location)
            flat(location.locations)
        return result

    flat(nested_locations)
    return result



from komora_syncer.connections.NetboxConnection import NetboxConnection

import logging
logger = logging.getLogger(__name__)

class NetboxBase:
    def __init__(self):
        self.netbox = NetboxConnection()
        self.netbox.open()


from komora_syncer.config import LoggingConfig
from komora_syncer.models.Synchronizer import Synchronizer

import os
import sys
import click
import traceback

from logging.config import fileConfig as logging_fileConfig
import logging
logger = logging.getLogger(__name__)


def setup_logger():
    try:
        log_conf = LoggingConfig()
        logging_fileConfig(log_conf._config, disable_existing_loggers=False)
    except FileNotFoundError as e:
        logger.error(f"File not found: {e}")
        logger.critical("Exiting")
        sys.exit()
    except Exception as e:
        raise e


@click.group()
def cli():
    pass


SYNC_OPTIONS = ['all', 'organizations', 'regions', 'sites', 'devices']
@cli.command()
@click.option('--sync', '-s', type=click.Choice(SYNC_OPTIONS), multiple=True, default=['all'], help="what should be synced")
@click.option('--configs', '-c', type=click.Path(), multiple=False, help="path to folder with configurations")
def synchronize(sync, configs):
    """
    Synchronize data between Netbox and Komora
    """

    # set custom configs folder path
    if configs:
        os.environ["CONFIG_FOLDER_PATH"] = os.path.abspath(configs)
    setup_logger()

    synchronizer = Synchronizer()

    if 'all' in sync or 'organizations' in sync:
        try:
            # Syncs Organizations / Tenants
            synchronizer.sync_organizations()
        except Exception as e:
            logger.error(f"Unable to synchronize organizations")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'regions' in sync:
        try:
            # Syncs Regions, Disctricts and Municipalities / Regions
            synchronizer.sync_regions()
        except Exception as e:
            logger.error(f"Unable to synchronize regions")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'sites' in sync:
        try:
            # Syncs Sites / Site, Location
            synchronizer.sync_sites()
        except Exception as e:
            logger.error(f"Unable to synchronize sites")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return

    if 'all' in sync or 'devices' in sync:
        try:
            synchronizer.sync_devices()
        except Exception as e:
            logger.error(f"Unable to synchronize devices")
            logger.debug(f"{e}\n{traceback.format_exc()}")
            logger.critical("Exiting")
            return


if __name__ == "__main__":
    # Display CLI
    cli()



def build_tenant_name(client_id, name):
    if client_id and name:
        return f"{client_id} - {name}"[0:100].strip()
    elif client_id:
        return f"{client_id}"
    else:
        return f"{name}"[0:100].strip()



from komora_syncer.models.netbox.NbLocation import NbLocation
from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.config import get_config

from slugify import slugify

import logging
logger = logging.getLogger(__name__)


class NbSite(NetboxBase):
    def __init__(self, komora_obj):
        NetboxBase.__init__(self)
        # Locations are already flatten via Komora site obj
        self.site_locations = komora_obj.flatten_locations

        self.name = komora_obj.name
        self.slug = slugify(self.name)

        self.description = komora_obj.fullName
        self.comments = komora_obj.description

        self.physical_address = komora_obj.address.text if komora_obj.address else ""
        self.shipping_address = ""
        self.latitude = float(
            komora_obj.latitude[0:9]) if komora_obj.latitude else None
        self.longitude = round(float(
            komora_obj.longitude[0:10]),6) if komora_obj.longitude else None

        self.komora_id = komora_obj.id
        self.komora_url = f"{get_config()['komora']['KOMORA_URL']}/app/site/{self.komora_id}"

        # site id
        self.code = komora_obj.code
        # Types: Připojný bod, virtuální bod, etc
        self.type_name = komora_obj.typeName
        self.type_id = komora_obj.typeId

        self.api_object = None
        self.region = None
        self.tenant = None

        try:
            if komora_obj.address:
                municipality_komora_id = komora_obj.address.municipalityId

                if municipality_komora_id:
                    self.region = self.netbox.connection.dcim.regions.get(
                        cf_komora_id=municipality_komora_id)
            elif komora_obj.regionId:
                self.region = self.netbox.connection.dcim.regions.get(cf_komora_id=komora_obj.regionId)
        except Exception as e:
            logger.exception(
                f"Unable to find region {komora_obj.address} for site {self.name}")

        try:
            komora_org_id = komora_obj.organizationId

            if komora_org_id:
                self.tenant = self.netbox.connection.tenancy.tenants.get(
                    cf_komora_id=komora_org_id)
        except Exception as e:
            logger.exception(
                f"Unable to find tenant {komora_obj.organization} for site {self.name}")

    def find(self):
        # 1. lookup by KOMORA_ID
        if self.komora_id:
            try:
                netbox_site = self.netbox.connection.dcim.sites.get(
                    cf_komora_id=self.komora_id)
                if netbox_site:
                    self.api_object = netbox_site
                    return self.api_object
            except Exception as e:
                logger.exception(
                    f"Unable to get site by komora_id: {self.komora_id}")

        # 2. Lookup by name, if komora id is not preseted
            # - log a problem, when the name exists, but komora_id was not found
        try:
            netbox_site = self.netbox.connection.dcim.sites.get(
                name__ie=self.name)
            if netbox_site:
                logger.warning(
                    f"komora_id: {str(self.komora_id)} was not found, but Site {self.name} already exists")
                self.api_object = netbox_site
                return self.api_object

        except Exception as e:
            logger.exception(f"Unable to get site by name: {self.name}")

        return self.api_object

    def create(self):
        try:
            params = self.get_params()
            netbox_site = self.netbox.connection.dcim.sites.create(params)

            logger.info("Site: %s created sucessfully", self.name)
            self.api_object = netbox_site
        except Exception as e:
            logger.exception("Unable to create netbox site: %s", self.name)
            raise e

        return self.api_object

    def update(self, nb_site):
        try:
            if nb_site.update(self.get_params()):
                self.api_object = nb_site
                logger.info(f"Site: {self.name} updated successfuly")
        except Exception as e:
            logger.exception(f"Unable to update site {self.name}")

    def find_or_create(self):
        self.find()
        if not self.api_object():
            self.create()

        return self.api_object

    def synchronize(self):
        site = self.find()

        if site:
            self.update(site)
        else:
            self.create()

        # sync location of site
        for location in self.site_locations:
            parent = next((parent for parent in self.site_locations if parent.id ==
                          location.parentId), None) if location.parentId else None

            nb_location = NbLocation(location, parent, self.api_object)
            nb_location.synchronize()

    def get_params(self):
        params = {
            "name": self.name,
            "slug": self.slug,
            "description": self.description,
            "comments": self.comments,
            "physical_address": self.physical_address,
            "shipping_address": self.shipping_address,
            "latitude": self.latitude,
            "longitude": self.longitude,
        }

        if self.komora_id:
            if type(params.get('custom_fields')) is dict:
                params['custom_fields']['komora_id'] = self.komora_id
                params['custom_fields']['komora_url'] = self.komora_url

                params['custom_fields']['code'] = self.code
                params['custom_fields']['type'] = self.type_name
            else:
                params['custom_fields'] = {"komora_id": self.komora_id,
                                           "komora_url": self.komora_url,
                                           "code": self.code,
                                           "type": self.type_name}

        if self.region:
            params['region'] = self.region.id

        if self.tenant:
            params['tenant'] = self.tenant.id

        return params



from komora_syncer.models.netbox.NetboxBase import NetboxBase
from komora_syncer.config import get_config
from komora_syncer.connections.NetboxConnection import NetboxConnection

import logging
logger = logging.getLogger(__name__)


class NbDevice(NetboxBase):
    def __init__(self, komora_obj):
        NetboxBase.__init__(self)
        self.name = komora_obj.name
        self.komora_id = komora_obj.id
        self.site_name = komora_obj.siteName
        self.komora_url = f"{get_config()['komora']['KOMORA_URL']}/app/device/{self.komora_id}"
        self.api_object = None

    def find(self):
        # 1. lookup Device by KOMORA_ID
        if self.komora_id:
            try:
                netbox_device = self.netbox.connection.dcim.devices.get(
                    cf_komora_id=self.komora_id)
                if netbox_device:
                    self.api_object = netbox_device
                    return self.api_object
            except Exception as e:
                logger.exception(
                    f"Unable to get Device by komora_id: {self.komora_id}")

        # 2. Lookup device by name, if komora id is not preseted
            # - log a problem, when the name exists, but komora_id was not found
        try:
            netbox_site = self.netbox.connection.dcim.sites.get(
                name__ie=self.site_name)
            netbox_device = self.netbox.connection.dcim.devices.get(
                name__ie=self.name, site_id=netbox_site.id)

            if netbox_device:
                logger.warning(
                    f"komora_id: {str(self.komora_id)} was not found, but Device {self.name} at site {self.site_name} already exists")
                self.api_object = netbox_device
                return self.api_object

        except Exception as e:
            logger.exception(f"Unable to get Device by name: {self.name}")

        return self.api_object

    def update(self, nb_device):
        try:
            if nb_device.update(self.get_params()):
                self.api_object = nb_device
                logger.info(f"Device: {self.name} updated successfuly")
        except Exception as e:
            logger.exception(f"Unable to update device {self.name}")

    def synchronize(self):
        device = self.find()

        if device:
            self.update(device)
        else:
            logger.info(
                f"Device {self.name} - komora_id: {self.komora_id} not found in netbox")

    def get_params(self):
        params = {}

        if self.api_object:
            if type(self.api_object.custom_fields) is dict:
                params['custom_fields'] = self.api_object.custom_fields
                params['custom_fields']['komora_id'] = self.komora_id
                params['custom_fields']['komora_url'] = self.komora_url
        else:
            params['custom_fields'] = {"komora_id": self.komora_id,
                                       "komora_url": self.komora_url}

        return params

    def get_nb_devices_data():
        import requests
        import json

        # TODO: WORKAROUND
        pynetbox = NetboxConnection()
        pynetbox.open()

        # FILTER OUT Passive Components, etc.
        # - device roles which aren't pass to Komora
        # TODO: move to config file
        filter_device_roles = pynetbox.connection.dcim.device_roles.filter(
            name=["Passive component", "unknown", "DWDM"])

        # json.dumps - IT will replace ' with " (single quotes aren't supported in graphql)
        filter_device_role_ids = json.dumps(
            [str(role.id) for role in filter_device_roles])

        # pass IDS to graphql query
        query = """
          {
            device_list(role_id__n: %s) {
              id,
              name,
              primary_ip4 {
                id,
                address
              },
              comments,
              serial,
              custom_fields,
              location {
                id,
                name,
                custom_fields
              },
              tenant {
                id,
                name,
                custom_fields
              }
              site {
                id,
                name,
                custom_fields,
                tenant{
                  id,
                  name,
                  custom_fields
                }
              },
              interfaces {
                id,
                name,
                description
              }
            }
          }
        """ % filter_device_role_ids

        try:
            url = get_config()['netbox']['NETBOX_GRAPHQL_URL']
            req = requests.post(url, json={'query': query}, headers={
                'Authorization': f"Token {get_config()['netbox']['NETBOX_API_TOKEN']}"})
            json_data = json.loads(req.text)
            return json_data
        except Exception as e:
            logger.exception(e)
            raise(e)
